//////////////////////////////////////////////////////////////////
// Symbolic
//
// Library for the construction of nested symbolic propositions
// The Flying Pig!
// Started 5/8/2003
// http://www.flypig.co.uk
// 
// 
// The MIT License
// See symbolic.h, COPYING file or website for licence
// 
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
// Includes

#include "symbolic.h"
#include "symbolic_private.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <math.h>
#include <float.h>

//////////////////////////////////////////////////////////////////
// Defines

//////////////////////////////////////////////////////////////////
// Structures

//////////////////////////////////////////////////////////////////
// Global variables

//////////////////////////////////////////////////////////////////
// Function prototypes

//////////////////////////////////////////////////////////////////
// Main application

//////////////////////////////////////////////////////////////////
// Recursively approximage an operation if possible
double ApproximateOperation (Operation * psOp) {
  double              fReturn = NAN;
  double              fVar1;
  double              fVar2;

  if (psOp) {
    switch (psOp->eOpType) {
      case OPTYPE_INTEGER:
        fReturn = (double)psOp->Vars.nInteger;
        break;
      case OPTYPE_TRUTHVALUE:
        fReturn = (double)psOp->Vars.boTruth;
        break;
      case OPTYPE_VARIABLE:
        if (psOp->Vars.psVar->psValue) {
          fReturn = GetVariable (psOp->Vars.psVar->psValue);
        }
        break;
      case OPTYPE_UNARY:
        fVar1 = ApproximateOperation (psOp->Vars.psUnary->psVar1);
        switch (psOp->Vars.psUnary->eOpType) {
          case OPUNARY_NEGATIVE:
            fReturn = - fVar1;
            break;
          case OPUNARY_NOT:
            fReturn = 1.0 - fVar1;
            break;
          case OPUNARY_COSEC:
            fReturn = 1.0 / sin (fVar1);
            break;
          case OPUNARY_SINH:
            fReturn = sinh (fVar1);
            break;
          case OPUNARY_COSH:
            fReturn = cosh (fVar1);
            break;
          case OPUNARY_TANH:
            fReturn = sinh (fVar1);
            break;
          case OPUNARY_COS:
            fReturn = cos (fVar1);
            break;
          case OPUNARY_SIN:
            fReturn = sin (fVar1);
            break;
          case OPUNARY_TAN:
            fReturn = tan (fVar1);
            break;
          case OPUNARY_SEC:
            fReturn = 1.0 / cos (fVar1);
            break;
          case OPUNARY_COT:
            fReturn = 1.0 / tan (fVar1);
            break;
          case OPUNARY_ARCSIN:
            fReturn = asin (fVar1);
            break;
          case OPUNARY_ARCCOS:
            fReturn = acos (fVar1);
            break;
          case OPUNARY_ARCTAN:
            fReturn = atan (fVar1);
            break;
          case OPUNARY_LN:
            fReturn = log (fVar1);
          case OPUNARY_EXP:
            fReturn = exp (fVar1);
            break;
          case OPUNARY_ABS:
            fReturn = fabs (fVar1);
            break;
          default:
            // Do nothing (return NAN)
            break;
        }
        break;
      case OPTYPE_BINARY:
        fVar1 = ApproximateOperation (psOp->Vars.psBinary->psVar1);
        fVar2 = ApproximateOperation (psOp->Vars.psBinary->psVar2);
        switch (psOp->Vars.psBinary->eOpType) {
          case OPBINARY_LAND:
            fReturn = (int)fVar1 && (int)fVar2;
            break;
          case OPBINARY_LOR:
            fReturn = (int)fVar1 || (int)fVar2;
            break;
          case OPBINARY_LIMP:
            break;
          case OPBINARY_POW:
            fReturn = pow (fVar1, fVar2);
            break;
          case OPBINARY_ADD:
            fReturn = fVar1 + fVar2;
            break;
          case OPBINARY_SUB:
            fReturn = fVar1 - fVar2;
            break;
          case OPBINARY_MUL:
            fReturn = fVar1 * fVar2;
            break;
          case OPBINARY_DIV:
            fReturn = (int)fVar1 / (int)fVar2;
            break;
          case OPBINARY_DIVIDE:
            fReturn = fVar1 / fVar2;
            break;
          case OPBINARY_MOD:
            if (!(isnan (fVar1) || isnan (fVar2))) {
              fReturn = fmod (fVar1, fVar2);
            }
            break;
          case OPBINARY_OR:
            fReturn = (int)fVar1 | (int)fVar2;
            break;
          case OPBINARY_AND:
            fReturn = (int)fVar1 & (int)fVar2;
            break;
          case OPBINARY_BIC:
            fReturn = (int)fVar1 & !(int)fVar2;
            break;
          case OPBINARY_EOR:
            fReturn = (int)fVar1 ^ (int)fVar2;
            break;
          case OPBINARY_ROR:
            break;
          case OPBINARY_ROL:
            break;
          case OPBINARY_EQ:
            if (!(isnan (fVar1) || isnan (fVar2))) {
              fReturn = (fVar1 == fVar2);
            }
            break;
          case OPBINARY_NE:
            if (!(isnan (fVar1) || isnan (fVar2))) {
              fReturn = (fVar1 != fVar2);
            }
            break;
          case OPBINARY_LT:
            if (!(isnan (fVar1) || isnan (fVar2))) {
              fReturn = (fVar1 < fVar2);
            }
            break;
          case OPBINARY_LE:
            if (!(isnan (fVar1) || isnan (fVar2))) {
              fReturn = (fVar1 <= fVar2);
            }
            break;
          case OPBINARY_GT:
            if (!(isnan (fVar1) || isnan (fVar2))) {
              fReturn = (fVar1 > fVar2);
            }
            break;
          case OPBINARY_GE:
            if (!(isnan (fVar1) || isnan (fVar2))) {
              fReturn = (fVar1 >= fVar2);
            }
            break;
          case OPBINARY_IND:
            break;
          default:
            break;
        }
        break;
      case OPTYPE_TERNARY:
        break;
      default:
        // Do nothing (returns NAN)
        break;
    }
  }
  return fReturn;
}
